package cz.jzitnik.game.mobs.services.pig;

import cz.jzitnik.game.Game;
import cz.jzitnik.game.annotations.EntityLogic;
import cz.jzitnik.game.annotations.EntitySpawn;
import cz.jzitnik.game.entities.Block;
import cz.jzitnik.game.entities.items.ItemBlockSupplier;
import cz.jzitnik.game.mobs.EntityLogicInterface;
import cz.jzitnik.game.mobs.EntityLogicProvider;
import cz.jzitnik.game.mobs.EntitySpawnInterface;
import cz.jzitnik.game.sprites.Pig;
import cz.jzitnik.tui.ScreenMovingCalculationProvider;
import org.jline.terminal.Terminal;

import java.util.*;

@EntitySpawn
@EntityLogic("pig")
public class PigLogic implements EntityLogicInterface, EntitySpawnInterface {
    private final Random random = new Random();

    @Override
    public void nextIteration(EntityLogicProvider.EntityLogicMobDTO entityLogicMobDTO) {
        int pigX = entityLogicMobDTO.getX();
        int pigY = entityLogicMobDTO.getY();
        var game = entityLogicMobDTO.getGame();
        var pig = entityLogicMobDTO.getMob();
        var world = game.getWorld();
        var pigData = (PigData) pig.getData();

        boolean updated = false;
        int newPigX = pigX;
        int newPigY = pigY;

        // Reduce movement cooldown
        if (pigData.getMovementCooldown() > 0) {
            pigData.setMovementCooldown(pigData.getMovementCooldown() - 1);
            return; // Skip movement this iteration
        }

        // Determine movement direction
        int direction = pigData.getLastDirection();
        if (random.nextInt(10) < 3) { // 30% chance to change direction
            direction = -direction;
        }
        pigData.setLastDirection(direction);

        // Update sprite direction
        if (direction == 1) {
            pig.setSpriteState(Pig.PigState.RIGHT);
        } else {
            pig.setSpriteState(Pig.PigState.LEFT);
        }

        List<Block> blocksAhead = world[pigY][pigX + direction];
        if (!game.isSolid(blocksAhead)) {
            world[pigY][pigX].remove(pig);
            world[pigY][pigX + direction].add(pig);
            newPigX = pigX + direction;
            updated = true;
            pigData.setJumpAttempts(0); // Reset jump attempts when moving forward
        } else {
            List<Block> blocksAboveAhead = world[pigY - 1][pigX + direction];
            List<Block> blocksTwoAboveAhead = world[pigY - 2][pigX + direction];

            // Jump if there is only one block height obstacle and limit jump attempts
            if (!game.isSolid(blocksAboveAhead) && game.isSolid(blocksAhead) && !game.isSolid(blocksTwoAboveAhead)) {
                if (pigData.getJumpAttempts() < 2) { // Limit jumping attempts to prevent infinite jumping
                    world[pigY][pigX].remove(pig);
                    world[pigY - 1][pigX + direction].add(pig);
                    newPigX = pigX + direction;
                    newPigY = pigY - 1;
                    updated = true;
                    pigData.setJumpAttempts(pigData.getJumpAttempts() + 1);
                }
            }
        }

        // Falling logic (avoid long falls)
        while (updated) {
            if (!game.isSolid(world[newPigY + 1][newPigX])) {
                if (newPigY - pigY < 3) { // Only fall if it's at most 2 blocks drop
                    world[newPigY][newPigX].remove(pig);
                    world[newPigY + 1][newPigX].add(pig);
                    newPigY++;
                } else {
                    updated = false;
                }
            } else {
                updated = false;
            }
        }

        // Apply movement cooldown to slow down movement
        pigData.setMovementCooldown(random.nextInt(3) + 1); // 1-3 iterations cooldown
    }

    @Override
    public void spawn(int playerX, int playerY, Game game, Terminal terminal) {
        // Cordinates where player can see
        int[] data = ScreenMovingCalculationProvider.calculate(playerX, playerY, terminal.getHeight(), terminal.getWidth(), game.getWorld()[0].length, game.getWorld().length);
        var world = game.getWorld();
        int startX = data[0];
        int endX = data[1];

        // Left side
        int lstartX = startX - 20;
        int lendX = startX - 5;
        int lstartY = playerY - 15;
        int lendY = playerY + 15;

        if (countPrasata(lstartX, lendX, lstartY, lendY, game) < 3 && random.nextInt(100) < 2) {
            var spawnLocations = pigCanSpawn(lstartX, lendX, playerY, game);
            if (!spawnLocations.isEmpty()) {
                System.out.println(spawnLocations.size());
                for (int i = 0; i < Math.min(4, spawnLocations.size()); i++) {
                    var randomLocation = getRandomEntry(spawnLocations);
                    int x = randomLocation.getKey();
                    int y = randomLocation.getValue();

                    world[y][x].add(ItemBlockSupplier.Mobs.pig());
                }
            }
        }

        // Right side
        int rstartX = endX + 5;
        int rendX = endX + 20;
        int rstartY = playerY - 15;
        int rendY = playerY + 15;

        if (countPrasata(rstartX, rendX, rstartY, rendY, game) < 3 && random.nextInt(100) < 2) {
            var spawnLocations = pigCanSpawn(rstartX, rendX, playerY, game);
            if (!spawnLocations.isEmpty()) {
                System.out.println(spawnLocations.size());
                for (int i = 0; i < Math.min(4, spawnLocations.size()); i++) {
                    var randomLocation = getRandomEntry(spawnLocations);
                    int x = randomLocation.getKey();
                    int y = randomLocation.getValue();

                    world[y][x].add(ItemBlockSupplier.Mobs.pig());
                }
            }
        }
    }

    public static <K, V> Map.Entry<K, V> getRandomEntry(HashMap<K, V> map) {
        List<Map.Entry<K, V>> entryList = new ArrayList<>(map.entrySet());
        Random random = new Random();
        return entryList.get(random.nextInt(entryList.size()));
    }

    private HashMap<Integer, Integer> pigCanSpawn(int startX, int endX, int playerY, Game game) {
        var map = new HashMap<Integer, Integer>();
        var world = game.getWorld();
        for (int x = startX; x <= endX; x++) {
            for (int y = Math.max(0, playerY - 30); y < Math.min(world.length, playerY + 30); y++) {
                if (world[y][x].stream().anyMatch(i -> i.getBlockId().equals("grass"))) {
                    map.put(x, y - 1);
                }
            }
        }

        return map;
    }

    private long countPrasata(int startX, int endX, int startY, int endY, Game game) {
        long pigAmount = 0;
        for (int y = startY; y <= endY; y++) {
            for (int x = startX; x <= endX; x++) {
                pigAmount += game.getWorld()[y][x].stream().filter(i -> i.getBlockId().equals("pig")).count();
            }
        }

        return pigAmount;
    }
}
